#define TRUE 1
#define FALSE 0

extern int tag_width;
extern int rob_size;

#define TAG_WIDTH tag_width
#define ROB_SIZE rob_size  // (2^TAG_WIDTH)
#define PRODUCERS 7        // No. of CDB producers
#define FUS 7              // No. of function units

int checkparam(const char *param);

// optimization tricks
// setting a trick-variable to one enables a certain optimization

extern int trick1, trick2;

