/***************************************************************
**  techno.h            technology header module              **
**  950801 TGR 0.1      (from SMM venus.h/motorola.h)         **
***************************************************************/

#ifndef TECHNO_H
#define TECHNO_H

#define MOTOROLA  0     /* Motorola H4C technology, 1993 */
#define VENUS     1     /* Venus Design System,   
                            (Arno, SMM PhD Thesis) */

extern int cinv, cand, cor, cnand, cnor;
extern int cxor, cxnor, cmux, cdriv, cff;
extern int cramcell, delta;
extern double romfac;
extern int DRAMfac;
extern int dinv, dand, dor, dnand, dnor;
extern int dxor, dxnor, dmux, ddriv, dff;

extern int cram(),  cram2(),  cram3(),  crom(),  cDRAM();
extern int dram(),  dram2(),  dram3(),  drom(),  dDRAM();

extern int   init_techno();     /* initialize technology
                                   parameters */
extern char* name_techno();     /* name: current technology */

#endif

